/**
 * An implementation of CipherSaber by Raph Levien, written 27 Apr 2000.
 *
 * This implementation is placed in the public domain.
 *
 * Please see http://ciphersaber.gurus.com/ for more information about
 * CipherSaber.
 *
 **/

#include <stdio.h>
#include <string.h>

typedef struct _RC4Ctx RC4Ctx;

struct _RC4Ctx {
  int x;
  int y;
  int perm[256];
};

int
usage (void)
{
  fprintf (stderr, "Usage: saber [-d] key < infile > outfile\n");
  return 1;
}

static void
rc4_init (RC4Ctx *ctx, const char *key, int key_size)
{
  int i, j;
  int tmp;

  for (i = 0; i < 256; i++)
    ctx->perm[i] = i;
  j = 0;
  for (i = 0; i < 256; i++)
    {
      j += ctx->perm[i] + key[i % key_size];
      j &= 255;
      tmp = ctx->perm[i];
      ctx->perm[i] = ctx->perm[j];
      ctx->perm[j] = tmp;
    }
#ifdef VERBOSE
  for (i = 0; i < 256; i++)
    {
      fprintf (stderr, "%d ", ctx->perm[i]);
    }
#endif
  ctx->x = 0;
  ctx->y = 0;
}

static char
rc4_byte (RC4Ctx *ctx)
{
  int x, y;
  int tmp, result;

  x = ctx->x;
  y = ctx->y;

  x = (x + 1) & 255;
  y = (y + ctx->perm[x]) & 255;
  tmp = ctx->perm[x];
  result = ctx->perm[y];
  ctx->perm[x] = result;
  ctx->perm[y] = tmp;
  result = ctx->perm[(result + tmp) & 255];

  ctx->x = x;
  ctx->y = y;
  return result;
}

int
main (int argc, char **argv)
{
  char *key = NULL;
  int key_len;
  int i;
  int decrypt = 0;
  RC4Ctx ctx;
  int n_bytes;
  char *tmpkey;
  int c;

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-d"))
	decrypt = 1;
      else if (key == NULL)
	key = argv[i];
      else
	return usage ();
    }
  if (key == NULL)
    return usage ();
  key_len = strlen (key);
  tmpkey = (char *)malloc (key_len + 10);
  if (decrypt)
    {
      n_bytes = fread (tmpkey + key_len, 1, 10, stdin);
    }
  else
    {
      FILE *f_rand;
      f_rand = fopen ("/dev/random", "rb");
      if (f_rand == NULL)
	{
	  fprintf (stderr, "Error opening /dev/random\n");
	  exit (1);
	}
      n_bytes = fread (tmpkey + key_len, 1, 10, f_rand);
      fclose (f_rand);
      fwrite (tmpkey + key_len, 1, 10, stdout);
    }
  if (n_bytes != 10)
    {
      fprintf (stderr, "Error reading IV\n");
      exit (1);
    }
  memcpy (tmpkey, key, key_len);
  rc4_init (&ctx, tmpkey, key_len + 10);
  memset (tmpkey, 0, key_len + 10);
  free (tmpkey);
  while ((c = getchar ()) != EOF)
    putchar (c ^ rc4_byte (&ctx));
  memset (&ctx, 0, sizeof (ctx));
  return 0;
}
