#include <stdio.h>
#define MAX_KEY_SZ 246
#define STATE_SZ 256
#define SUGG_MAX_KEY_SZ 54
#define SUGG_MIN_KEY_SZ 8   /* not enforced, just recommended */
#define IV_SZ 10

unsigned char key[MAX_KEY_SZ + IV_SZ + 1];
int key_len = 0, decrypt = 0;
unsigned char iv[IV_SZ + 1];
unsigned char state[MAX_KEY_SZ + 1];
unsigned char A, B;

int usage() {
  fprintf(stderr, "usage: ciphersaber [-d] key < infile > outfile\n");
  return 0;
}

int rc4_init() {
  unsigned int i;
  unsigned int N;
  unsigned int tmp;

  for (i = 0; i < MAX_KEY_SZ; i++) {
    state[i] = i;
  }
  A = 0;
  B = 0;
  for (N = 0; N < STATE_SZ; N++) {
    B += key[A] + state[A] % 256;
    tmp = state[A];
    state[A] = state[B];
    state[B] = tmp;
    A++ % key_len;
  }
  A = 0;
  B = 0;

  return 1;
}

unsigned char rc4_byte() {
  unsigned int index;
  
  A++ % STATE_SZ;
  B += state[A] + state[B] % STATE_SZ;
  index = state[A] + state[B] % STATE_SZ;
  return state[index] % STATE_SZ;
}

int main(int argc, char* argv[]) {
  unsigned int i;  
  int c;

  if ( ! (argc > 1)) {
    return usage();
  }
  if ( ! strcmp(argv[1], "-d")) {
    decrypt = 1;
    strncpy(key, argv[2], MAX_KEY_SZ);
  }
  else {
    decrypt = 0;
    strncpy(key, argv[1], MAX_KEY_SZ);
  }
  if (key == 0) {
    return usage();
  }
  key_len = strlen(key);

  if (key_len < SUGG_MIN_KEY_SZ) {
   fprintf(stderr, "WARNING:  password only has %d characters.  We recommend the password to be at least %d characters long.\n", key_len, SUGG_MIN_KEY_SZ);
  }
  if (key_len > SUGG_MAX_KEY_SZ) {
    fprintf(stderr, "WARNING: ironically, passord has to many characters (%d).  It is recommended not to use more than %d characters so that IV get's mixed in properly during the key setup.\n", key_len, SUGG_MAX_KEY_SZ);
  }

  /* append 10 bytes of randomness to key */
  if (decrypt) {
    /* read IV from begining of file */
    if ( IV_SZ > fread(iv, IV_SZ, 1, stdin)) {
      fprintf(stderr, "file doesn't have %d characters for IV\n", IV_SZ);
    }
  }
  else {
    FILE *f_rand;
    f_rand = fopen("/dev/random", "rb");
    if (f_rand == NULL) {
      fprintf (stderr, "Error opening /dev/random\n");
      exit (1);
    }
    
    for (i = 0; i < IV_SZ; i++) {
      iv[i] = (unsigned char)(getc(f_rand) % 256);
      putchar(iv[i]);
    }
    fclose(f_rand);
  }
  strncat(key, iv, IV_SZ);
  key_len = key_len + IV_SZ;
  
  rc4_init();

  while ((c = getchar()) != EOF) {
    putchar(c^rc4_byte());
  }

  /* kill all evidence */

  fprintf(stderr, "key: %s\n", key);
  fprintf(stderr, "decrypt: %d\n", decrypt);
  fprintf(stderr, "key length: %d\n", key_len);

  return 0;
}
