/* -*- Mode: C; c-file-style: "bsd" -*- */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include "cdb.h"

#define byte unsigned char

/* compile with:

gcc -O6 -I./cdb csvec.c -o csvec cdb/cdb.a cdb/byte.a cdb/unix.a

*/

int iv_len = 0;			/* default to RC4 */
int ROUNDS = 1;

int PROGRESS = 0;
int PROGRESS_LIM = 10000;
int VERBOSE = 0;
int DEBUG = 0;
int TRACE = 0;

int precompute = 0;

byte *word[500000];

typedef struct rc4_key
{
    byte state[256];
    byte x;
    byte y;
} rc4_key;

void prepare_key(byte *key_data_ptr, int key_data_len, rc4_key *key);
void rc4(byte *buffer_ptr, int buffer_len, rc4_key *key);

void xor_precompute( int argc, char *argv[] );
int cs_recur( char* pass, int word_num, int *tries );
int rc4_recur( char* pass, int word_num );

void cdb_start( void );

void farena_reset( void );
void farena_free( void );
void funalloc( void );
void *falloc( int len );
byte *fdup( const byte *s, int len );
int fstrlen( const byte *s );
byte* fgetstr( FILE *f );

FILE *f, *out[20], *key[20];
char cdb_dir[256];
int p = 0, nwords;
int num_words;

int min_key_len;
int min_key_word_len, max_key_word_len;
int a_min_key_word_len[100], a_max_key_word_len[100];
int min_vector_word_len, max_vector_word_len;

int first_word_of_len[100];

int usage( void ) {
    fprintf( stderr, "usage:\n\n"
"encrypt [-c n][-i n][-r n][-h|-v|-d|-p] d n [kl_r [...]] vl_r dict [...]\n\n"
"or\n\n"
"encrypt -x [-h|-v|-d|-p] words [...] | cdbmake pair.N.cdb tmp.N.cdb\n\n"
"encrypt options:\n\n"
"\t-c n\t\t0 = RC4, 1,2 are ciphersaber1,2; default is RC4\n"
"\t-i n\t\tnon-stanard choice of ciphersaber iv length\n"
"\t-r n\t\tnon-standard choice of key-sched rounds\n"
"\t-v\t\tverbose output\n"
"\t-d\t\tdebugging output\n"
"\t-t\t\ttracing output\n"
"\t-p\t\tshow progress output every n searches\n"
"\t-P n\t\tchange progress output to every n searches\n"
"\td\t\tthe directory where precomputed word pair xors are\n"
"\tn\t\tnumber of words to make the key\n"
"\tkl_r [...]\tranges of key lengths, there should be nwords\n"
"\t\t\tof ranges, one for each word specified in nwords\n"
"\tvl_r\t\trange of vector length words\n"
"\tdict [...]\tfiles full of words (note the concatenation of the\n"
"\t\t\tfiles must be form a list of words of strictly\n"
"\t\t\tincreasing length)\n\n"
"(ranges are written n for fixed length n, or n-m for [n,m])\n\n"
"encrypt -x options:\n\n"
"\t-x\t\tprecompute word pair xors and output in cdbmake\n"
"\t\t\tformat (all words must be same length as pair.N.cdb\n"
"\t\t\tholds only words of length N to reduce storage\n"
"\twords [...]\tfiles of words of length N\n\n" );
    exit( EXIT_FAILURE );
}

void read_pair( const char *arg, int *min, int *max ) {
    char copy[256], *ptr;

    strncpy( copy, arg, 256 );
    *min = 0;
    *max = 0;
    ptr = strchr( copy, '-' );
    if ( ptr == 0 ) { *min = *max = atoi( copy ); return; }
    *ptr++ = '\0';
    *min = atoi( copy );
    *max = atoi( ptr );
}

/* sort by length and then value */

int flstrcmp( const void* a, const void* b ) {
    byte *A = *(byte**)a-1, *B = *(byte**)b-1;
    return strcmp( A, B );
}

/* sort by value */

int fstrcmp( const void* a, const void* b ) {
    byte *A = *(byte**)a, *B = *(byte**)b;
    return strcmp( A, B );
}

int main( int argc, char *argv[] ) {
    int i, j, a=1, args, prev_word_len, curr_word_len, c, tries;
    byte pass[1000];

    while ( (c = getopt(argc, argv, "hxi:c:r:vdtpP:")) != -1 ) {
	switch (c) {
	case 'h': usage(); break;
	case 'i': iv_len = atoi( optarg ); 
	    if ( iv_len == 1 ) {
		fprintf( stderr, "error: iv length can not be 1, "
			 "must be either 0 or larger\n" );
		exit( EXIT_FAILURE );
	    }
	    break;
	case 'v': VERBOSE = 1; break;
	case 'd': DEBUG = 1; break;
	case 't': TRACE = 1; break;
	case 'P': PROGRESS_LIM = atoi( optarg ); /* and fall through */
	case 'p': PROGRESS = 1; break;
	case 'r': ROUNDS = atoi( optarg ); break;
	case 'x': precompute = 1; break;
	case 'c': 
	    switch ( atoi( optarg ) ) {
	    case 0: /* RC4 mode same with iv_len = 0 */
		iv_len = 0; ROUNDS = 1;	break;
	    case 1: /* CS1 */
		iv_len = 10; ROUNDS = 1; break;
	    case 2: /* CS2 same as CS1 but with more rounds */
		iv_len = 10; ROUNDS = 20; break;
	    default: usage();
	    }
	    break;
	case '?':
	    fprintf( stderr, "error: unkown option '-%c'\n", optopt );
	    usage();
	    break;
	case ':':
	    fprintf( stderr, "error: missing '-%c' option argument\n", 
		     optopt );
	    usage();
	    break;
	}
    }
	    
    args = argc-1;
    a = optind;

    if ( precompute ) {
	xor_precompute( argc-optind, argv+a );
    }

    if ( args < 2 ) { usage(); }

    strcpy( cdb_dir, argv[a++] );

    num_words = atoi( argv[a++] );

    if ( args < 4+num_words ) { usage(); }

    min_key_word_len = 256;
    max_key_word_len = 0;
    for ( i = 1; i <= num_words; i++ ) {
	read_pair( argv[a++], &a_min_key_word_len[i], &a_max_key_word_len[i] );
	if ( a_min_key_word_len[i] < min_key_word_len ) {
	    min_key_word_len = a_min_key_word_len[i];
	}
	if ( a_max_key_word_len[i] > max_key_word_len ) {
	    max_key_word_len = a_max_key_word_len[i];
	}
    }

    read_pair( argv[a++], &min_vector_word_len, &max_vector_word_len );

    cdb_start();

    /* sanity checking */

    if ( iv_len && num_words < 2 ) {
	fprintf( stderr, "error: need a min of 2 words if you have an IV\n" );
	exit( EXIT_FAILURE );
    }

    /* read all the words in */

    nwords = 0;
    for ( ; a <= args; a++ ) {
	if ( VERBOSE ) {
	    fprintf( stderr, "reading file %s\n", argv[a] );
	}
	f = fopen( argv[a], "r" );
	if ( !f ) { 
	    fprintf( stderr, "warning: can't open %s\n", argv[a] ); 
	    continue; 
	}
	
	while ( !feof(f) ) {
	    if ( ( word[nwords] = fgetstr( f ) ) ) {
		curr_word_len = fstrlen( word[nwords] );
		if ( curr_word_len >= 100 ) {
		    fprintf( stderr, "error: words must be < 100 chars\n" );
		    exit( EXIT_FAILURE );
		}
/* throw away words which are too short or too long */
		if ( curr_word_len >= min_key_word_len && 
		     curr_word_len <= max_key_word_len ) {
		    if ( DEBUG ) {
			fprintf( stderr, "word[%d] = \"%s\"\n", nwords, word[nwords] );
		    }
		    nwords++;
		} else {
		    funalloc();
		}
	    }
	}
	fclose( f );
    }
    if ( nwords == 0 ) { 
	fprintf( stderr, "error: no words found\n" ); 
	exit( EXIT_FAILURE ); 
    }

    fprintf( stderr, "selected %d words of length [%d-%d]\n", 
	     nwords, min_key_word_len, max_key_word_len );

/* sort them into increasing length order 
   and secondarily alphabetic order */

    if ( VERBOSE ) {
	fprintf( stderr, "sorting...\n" );
    }
    if ( DEBUG ) {
	fprintf( stderr, "before sorting:\n" );
	for ( i = 0; i < nwords; i++ ) {
	    fprintf( stderr, "word[%d] = \"%s\"\n", i, word[i] );
	}
    }

    qsort( word, nwords, sizeof( byte* ), flstrcmp );
    if ( VERBOSE ) {
	fprintf( stderr, "done sorting\n" );
    }

    if ( DEBUG ) {
	fprintf( stderr, "after sorting:\n" );
	for ( i = 0; i < nwords; i++ ) {
	    fprintf( stderr, "word[%d] = \"%s\"\n", i, word[i] );
	}
    }

/* find start points for word length boundaries */
    for ( i = 0; i < max_key_word_len; i++ ) {
	first_word_of_len[i] = 0;
    }

    prev_word_len = 0;
    curr_word_len = 0;
    for ( i = 0; i < nwords; i++ ) {
	curr_word_len = fstrlen( word[i] );

	if ( curr_word_len > prev_word_len ) {
	    if ( prev_word_len > 0 ) {
		for ( j = prev_word_len+1; j <= curr_word_len; j++ ) {
		    first_word_of_len[j] = i;
		    if ( DEBUG ) {
			fprintf( stderr, "setting first[%d] = %d\n", j, i );
		    }
		}
	    }
	    prev_word_len = curr_word_len;
	}
    }

    for ( i = curr_word_len+1; i <= max_key_word_len+1; i++ ) {
	first_word_of_len[i] = nwords;
    }

    if ( DEBUG ) {
	fprintf( stderr, "first word of len skip ahead table:\n" );
	for ( i = min_key_word_len; i <= max_key_word_len+1; i++ ) {
	    fprintf( stderr, "first[%d] = %d\n", i, first_word_of_len[i] );
	}
    }

    if ( VERBOSE ) {
	fprintf( stderr, "done reading words\n" );
    }

    pass[0] = '\0';
    if ( iv_len == 0 ) {
	rc4_recur( pass, 0 );
    } else {
	min_key_len = a_min_key_word_len[1] + iv_len;
	tries = 0;
	cs_recur( pass, 0, &tries );
	if ( DEBUG ) {
	    fprintf( stderr, "examined %d strings\n", tries );
	}
    }
    exit( EXIT_SUCCESS );
}

struct cdb cdb[20];

void cdb_start( void ) {
    char cdb_file[256];
    FILE* f;
    int i;
    int fd;

    for ( i = min_vector_word_len; i <= max_vector_word_len; i++ ) {
	sprintf( cdb_file, "%s/pair.%d.cdb", cdb_dir, i );
	if ( VERBOSE ) {
	    fprintf( stderr, "opening \"%s\"...", cdb_file );
	}
	f = fopen( cdb_file, "r" );
	if (f == 0) { 
	    if ( VERBOSE ) {
		fprintf( stderr, "failed\n" ); 
	    }
	    continue; 
	}
	fd = fileno( f );
	cdb_init( &(cdb[i]), fd );
	if ( VERBOSE ) {
	    fprintf( stderr, "ok\n" );
	}
    }
}

byte words[1024*1024];

int cdb_lookup( struct cdb* c, char* k, int klen, char* pass ) {
    int found = 0, dlen, dpos, len;
    char save;

    cdb_findstart( c );
    while ( cdb_findnext( c, k, klen ) > 0 ) {
	if ( found == 0 ) {
	    if ( iv_len ) {
		len = strlen( pass );
		save = pass[len-iv_len]; pass[len-iv_len] = '\0';
		fprintf( stdout, "k=\"%s\",iv=\"%c%s\": ", 
			 pass, save, pass+(len-iv_len)+1 );
		pass[len-iv_len] = save;
	    } else {
		fprintf( stdout, "k=\"%s\": ", pass );
	    }
	} else {
	    fprintf( stdout, ", " );
	}
	found++;
	dlen = cdb_datalen( c ); 
	dpos = cdb_datapos( c );
	cdb_read( c, words, dlen, dpos );
	fwrite( words, dlen, 1, stdout );
    }
    if ( found > 0 ) { fputc( '\n', stdout ); fflush( stdout ); }
    return found;
}

int w[20];
int wordy;

const char* indent( void ) {
    static char tmp[ 30 ], res[ 100 ];
    int this_min, this_max, windex;
    int i, l;
    
    res[0] = '\0';
    for ( i = 1; i <= wordy; i++ ) {
	this_min = a_min_key_word_len[i];
	this_max = a_max_key_word_len[i];
	windex = w[i];
	sprintf( tmp, "[%d-%d=%d]", this_min, this_max, windex );
	strcat( res, tmp );
    }

    l = strlen(res);
    for ( i = 0; i < num_words*7+10-l; i++ ) {
	res[i+l] = '.';
    }
    res[i+l] = '\0';
    return res;
}

void rc4_try_key( char* pass ) {
    int i, xor_len, start_pass_len = strlen( pass );
    byte xor[22], save;
    rc4_key ks;

    if ( TRACE ) {
	fprintf( stderr, "%src4_tryk( \"%s\" )\n", indent(), pass );
    }
    if ( PROGRESS ) {
	if ( ++p > PROGRESS_LIM ) { p = 0; fputc( '.', stderr ); }
    }
    if ( VERBOSE ) {
	fprintf( stderr, "trying \"%s\"\n", pass );
    }
    prepare_key( pass, start_pass_len, &ks );
    memset( xor, 0, max_vector_word_len );
    rc4( xor, max_vector_word_len, &ks );
    for ( i = 0; i < max_vector_word_len; i++ ) {
	if ( xor[i] == '\0' || xor[i] == '\n' || xor[i] > 127 ) {
	    return;
	}
	xor_len = i+1;
	if ( xor_len >= min_vector_word_len ) {
	    save = xor[xor_len]; xor[xor_len] = '\0';
	    cdb_lookup( &(cdb[xor_len]), xor, xor_len, pass );
	    xor[xor_len] = save;
	}
    }
}

#define REQUEST_LONGER 1
#define DONE 0

int rc4_recur( char* pass, int word_num ) {
    int i, start_pass_len = strlen( pass );
    int this_min, this_max, first, last;
    
    word_num++;
    this_min = a_min_key_word_len[word_num];
    this_max = a_max_key_word_len[word_num];
    first = first_word_of_len[ this_min ];
    last = first_word_of_len[ this_max+1 ];
    for ( i = first; i < last; i++ ) {
	if ( DEBUG ) {
	    fprintf( stderr, "word %d searching lengths %d-%d at offsets %d-%d\n",
		     word_num, this_min, this_max, first, last-1 );
	}
	pass[start_pass_len] = '\0';
	strcpy( pass+start_pass_len, word[i] );
	if ( word_num == num_words ) {
	    rc4_try_key( pass );
	} else {
	    strcat( pass+start_pass_len, " " );
	    rc4_recur( pass, word_num );
	}
    }
    return DONE;
}

#if defined( USE_MACROS )

/* use macro version instead */
#define word_boundary( off ) (*(off) != ' ' && (*((off)-1) == ' '))

/* use macro version instead */
#define copy_and_remove_space( cpy, pass, boundary ) 	\
    memcpy( (copy), (pass), (boundary)-1 );		\
    strcpy( (copy)+(boundary)-1, (pass)+(boundary) )
#else

#endif

int word_boundary( char* off ) {
    return ( *off != ' ' && *(off-1) == ' ' );
}

int copy_and_remove_space( char *copy, const char *pass, int boundary ) {
/*    if ( TRACE ) { 
	fprintf( stderr, "%scpyremsp( \"%s\" )\n", indent(), pass );
    }
*/
    memcpy( copy, pass, boundary-1 );
    strcpy( copy+boundary-1, pass+boundary );
}

/* this one needs pruning */

char *why;

#define SKIP_TO_NEXT_LONGER_WORD( i )				\
    if ( TRACE ) {						\
        fprintf( stderr, "%sSKIP_NXT( \"%s\" ) = %s\n", 	\
		 indent(), pass, why );				\
    }								\
    i = first_word_of_len[ fstrlen( word[i] )+1 ]-1;            \
    continue

int cs_recur( char* pass, int word_num, int *sum_tries ) {
    int i, start_pass_len = strlen( pass );
    int this_min, this_max, first, last;
    int len, boundary, tries = 0;
    char copy[256];
    
    wordy = word_num;
    if ( TRACE ) {
	fprintf( stderr, "%scs_recur( \"%s\", %d )\n", indent(), pass, word_num );
    }
    word_num++;
    wordy = word_num;
    this_min = a_min_key_word_len[word_num];
    this_max = a_max_key_word_len[word_num];
    first = first_word_of_len[ this_min ];
    last = first_word_of_len[ this_max+1 ];
    for ( i = first; i < last; i++ ) {
	w[wordy] = fstrlen(word[i]);
	if ( DEBUG ) {
	    fprintf( stderr, "word %d searching lengths %d-%d at offsets %d-%d\n",
		     word_num, this_min, this_max, first, last-1 );
	}
	pass[start_pass_len] = '\0';
	strcpy( pass+start_pass_len, word[i] );
	strcat( pass+start_pass_len, " " );

	if ( word_num == num_words ) {
	    len = start_pass_len + fstrlen( word[i] )+1;
/* we're going to remove a space if this works out so don't count it */
	    if ( len-1 < min_key_len ) { 
		why = "key too short";
		SKIP_TO_NEXT_LONGER_WORD( i );
	    }
	    boundary = len-iv_len;
	    if ( !word_boundary( pass+boundary ) ) {
		why = "not word boundary";
		SKIP_TO_NEXT_LONGER_WORD( i );
	    }
/* remove the space before the word boundary */
	    copy_and_remove_space( copy, pass, boundary );
	    tries++;
	    rc4_try_key( copy );
	} else {
	    if ( cs_recur( pass, word_num, &tries ) == REQUEST_LONGER ) {
		wordy = word_num;
		why = "told by recursion";
		SKIP_TO_NEXT_LONGER_WORD( i );
	    }
	    wordy = word_num;
	}
    }
/* if we and our recursive calls collectively didn't try any strings
   at all advise the recursion level up one that we need a longer word
   because if we use a start of the same length we'll get the same
   result */
    
    if ( TRACE ) {
	fprintf( stderr, "%sreturning %s\n", indent(), tries>0 ? "DONE ALL" : "DID NOTHING" );
    }

    (*sum_tries) += tries;
    return (tries>0) ? DONE : REQUEST_LONGER;
}

/* pre-compute xored pairs of words of equal length and output
   in cdbmake format */

void xor_precompute( int argc, char *argv[] ) {
    int i, j, k, p = 0, nwords, curr_len = 0, prev_len = -1, a;
    char xor[100];
    FILE *f;

    for ( a = 0; a < argc; a++ ) {

	f = fopen( argv[a], "r" );
	if ( !f ) { 
	    fprintf( stderr, "warning: can't open %s\n", argv[a] ); 
	    continue; 
	}
	nwords = 0;
	while ( !feof(f) ) {
	    if ( ( word[nwords] = fgetstr( f ) ) ) {
		curr_len = fstrlen( word[nwords] );
		if ( curr_len >= 100 ) {
		    fprintf( stderr, "error: words must be < 100 chars\n" );
		    exit( EXIT_FAILURE );
		}
		if ( prev_len > -1 && curr_len != prev_len ) {
		    fprintf( stderr, "error: words must all be same length\n" );
		    exit( EXIT_FAILURE );
		}
		prev_len = curr_len;
		nwords++;
	    }
	}
	fclose( f );
	
	fprintf( stderr, "\nread %d words of length %d\n", nwords, curr_len );

/* do only upper diagonal to avoid swapped duplicates */ 

	xor[curr_len] = '\0';
	for ( i = 0; i < nwords-1; i++ ) {
	    for ( j = i+1; j < nwords; j++ ) {
		if ( PROGRESS ) {
		    if ( ++p > PROGRESS_LIM ) { p = 0; fputc( '.', stderr ); }
		}
		for ( k = 0; k < curr_len; k++ ) {
		    xor[k] = word[i][k] ^ word[j][k];
		    if ( xor[k] == '\0' || xor[k] == '\n' ) { goto skip; }
		}
		fprintf( stdout, "+%d,%d:%s->%s %s\n",
			 curr_len, 2*curr_len+1, xor, word[i], word[j] );
	    skip:
	    }
	}

/* reset arena -- re-use that memory in one swoop */

	farena_reset();

    }
    fputc( '\n', stdout );

/* don't bother freeing arena -- we're exiting */

    exit( 0 );
}

#define swap_byte(x,y) t = *(x); *(x) = *(y); *(y) = t

void prepare_key(byte *key_data_ptr, int key_data_len, rc4_key *key)
{
    int k;
    byte t;
    byte index1;
    byte index2;
    byte* state;
    short counter;

    state = &key->state[0];
    for(counter = 0; counter < 256; counter++)
	state[counter] = counter;
    key->x = 0;
    key->y = 0;
    index2 = 0;
    for ( k = 0; k < ROUNDS; k++ ) 
    {
	index1 = 0;
	for(counter = 0; counter < 256; counter++)
	{
	    index2 = (key_data_ptr[index1] + state[counter] + index2) % 256;
	    swap_byte(&state[counter], &state[index2]);
	    index1 = (index1 + 1) % key_data_len;
	}
    }
}

void rc4(byte *buffer_ptr, int buffer_len, rc4_key *key)
{
    byte t;
    byte x;
    byte y;
    byte* state;
    byte xorIndex;
    short counter;

    x = key->x;
    y = key->y;
    state = &key->state[0];
    for(counter = 0; counter < buffer_len; counter++)
    {
	x = (x + 1) % 256;
	y = (state[x] + y) % 256;
	swap_byte(&state[x], &state[y]);
	xorIndex = (state[x] + state[y]) % 256;
	buffer_ptr[counter] ^= state[xorIndex];
    }
    key->x = x;
    key->y = y;
}

#define MAX_STR_LEN 255
#define ARENA_SIZE (1024*1024)
byte *arena_prev = 0;
int arena_free = 0;

#define MAX_ARENA_STACK 100
byte *arena[MAX_ARENA_STACK]; /* zero initialised */
int arena_sp = -1;

void farena_reset( void ) {
    arena_sp = -1;
}

void farena_free( void ) {
    int i;
    for ( i = 0; i <= arena_sp; i++ ) {
	if ( arena[i] != 0 ) {
	    free( arena[i] );
	    arena[i] = 0;
	}
    }
    arena_sp = -1;
}

void *falloc( int len ) {
    void *ret;
    assert( len < ARENA_SIZE  );
    if ( arena_sp == -1 || len > arena_free ) { 
	arena_sp++;
	assert( arena_sp < MAX_ARENA_STACK );
	if ( arena[arena_sp] == 0 ) {
	    arena[arena_sp] = malloc( arena_free = ARENA_SIZE );
	}
    }
    ret = arena_prev = arena[arena_sp];
    arena_free -= len; arena[arena_sp] += len;
    return ret;
}

void funalloc( void ) {
    assert( arena[arena_sp] != 0 );
    arena_free -= (arena[arena_sp] - arena_prev);
    arena[arena_sp] = arena_prev;
}

byte *fdup( const byte *s, int len ) {
    byte *ret;
    assert( len <= MAX_STR_LEN );
    ret = falloc( len+2 );
    *ret++ = len;
    strcpy( ret, s );
    return ret;
}

int fstrlen( const byte *s ) {
    return s[-1];
}

byte* fgetstr( FILE *f ) {
    int len;
    byte str[MAX_STR_LEN];
    
    str[0] = '\0';
    if ( fgets( str, MAX_STR_LEN, f ) == 0 ) {
	return 0;
    }
    len = strlen( str );
    if ( str[len-1] == '\n' ) { str[len-1] = '\0'; len--; }
    if ( str[len-1] == '\r' ) { str[len-1] = '\0'; len--; }
    if ( str[len-1] == '\n' ) { str[len-1] = '\0'; len--; }
    return fdup( str, len );
}
